
#ifndef PLS_ALIGNED_STACK_H
#define PLS_ALIGNED_STACK_H

#include <cstdint>
#include <cstdlib>

#include "pls/internal/base/error_handling.h"

namespace pls {
    namespace internal {
        namespace base {
            class aligned_stack {
                // Keep bounds of our memory block
                char* memory_start_;
                char* memory_end_;

                // Current head will always be aligned to cache lines
                char* head_;

                static std::uintptr_t next_alignment(std::uintptr_t size);
                static char* next_alignment(char* pointer);
            public:
                typedef char* state;

                aligned_stack(): memory_start_{nullptr}, memory_end_{nullptr}, head_{nullptr} {};

                aligned_stack(char* memory_region, const std::size_t size):
                    memory_start_{memory_region},
                    memory_end_{memory_region + size},
                    head_{next_alignment(memory_start_)} {}

                template<typename T>
                T* push(const T& object) {
                    // Copy-Construct into desired memory location
                    return new (push<T>())T(object);
                }

                template<typename T>
                T* push() {
                    T* result = reinterpret_cast<T*>(head_);

                    // Move head to next aligned position after new object
                    head_ = next_alignment(head_ + sizeof(T));
                    if (head_ >= memory_end_) {
                        PLS_ERROR("Tried to allocate object on alligned_stack without sufficient memory!");
                    }

                    return result;
                }

                template<typename T>
                T pop() {
                    head_ = head_ - next_alignment(sizeof(T));

                    return *reinterpret_cast<T*>(head_);
                }

                state save_state() {
                    return head_;
                }

                void reset_state(state new_state) {
                    head_ = new_state;
                }
            };
        }
    }
}

#endif //PLS_ALIGNED_STACK_H
